/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "../List.h"

#include "lp.h"
#include "error.h"
/*-------------- Public functions -------------------*/
int getport(char *rhost);
char *getline(FILE *cfp);
int getq(struct queue *(*namelist[]), char *SpoolingDir);
int checkremote();
void ParseData(List* Listing, char *data, int i);
/*-------------- Private functions -------------------*/
static int compar(register struct queue **p1, register struct queue **p2);

/*
 * Create a connection to the remote printer server.
 * Most of this code comes from rcmd.c.
 */
int getport(char *rhost)
{
	struct hostent *hp;
	struct servent *sp;
	struct sockaddr_in sin;
	int s, timo = 1, lport = IPPORT_RESERVED - 1;
	int err;

	/*
	 * Get the host address and port number to connect to.
	 */
	if (rhost == NULL)
	{		
		fatal("No remote host to connect to.\n");
		return -1;
	}	
	hp = gethostbyname(rhost);
	if (hp == NULL)
	{
		char msg[MAXPATHLEN];
		sprintf(msg, "Host %s is unknown\n", rhost);
		fatal(msg);
		return -1;
	}	
	sp = getservbyname("printer", "tcp");
	if (sp == NULL)
	{  		
		fatal("Unknown service: printer/tcp\n");
		return -1;
	}	
	bzero((char *)&sin, sizeof(sin));
	bcopy(hp->h_addr, (caddr_t)&sin.sin_addr, hp->h_length);
	sin.sin_family = hp->h_addrtype;
	sin.sin_port = sp->s_port;

	/*
	 * Try connecting to the server.
	 */
retry:
	s = rresvport(&lport);
	if (s < 0)
		return(-1);
	if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		err = errno;
		(void) close(s);
		errno = err;
		if (errno == EADDRINUSE) {
			lport--;
			goto retry;
		}
		if (errno == ECONNREFUSED && timo <= 16) {
			sleep(timo);
			timo *= 2;
			goto retry;
		}
		return(-1);
	}
	return(s);
}

/*
 * Getline reads a line from the control file cfp, removes tabs, converts
 *  new-line to null and leaves it in line.
 * Returns NULL at EOF or the line read.
 */
char *getline(FILE *cfp)
{
	static char line[BUFSIZ];
	int linel = 0;
	char *lp = line;
	register c;

	while ((c = getc(cfp)) != '\n') 
	{
		if (c == EOF)
			return NULL;
		if (c == '\t') 
		{
			do 
			{
				*lp++ = ' ';
				linel++;
			}			
			while ((linel & 07) != 0);
			continue;
		}
		*lp++ = c;
		linel++;
	}
	*lp++ = '\0';
	return line;
}

/*
 * Scan the current directory and make a list of daemon files sorted by
 * creation time.
 * Return the number of entries and a pointer to the list.
 */
int getq(struct queue *(*namelist[]), char* SpoolingDir)
{
	register struct direct *d;
	register struct queue *q, **queue;
	register int nitems;
	struct stat stbuf;
	DIR *dirp;
	int arraysz;
	static int compar();

	if ((dirp = opendir(SpoolingDir)) == NULL)
		return(-1);
	if (fstat(dirp->dd_fd, &stbuf) < 0)
		goto errdone;

	/*
	 * Estimate the array size by taking the size of the directory file
	 * and dividing it by a multiple of the minimum size entry. 
	 */
	arraysz = (stbuf.st_size / 24);
	queue = (struct queue **)malloc(arraysz * sizeof(struct queue *));
	if (queue == NULL)
		goto errdone;

	nitems = 0;
	while ((d = readdir(dirp)) != NULL) {
		if (d->d_name[0] != 'c' || d->d_name[1] != 'f')
			continue;	/* daemon control files only */
		if (stat(d->d_name, &stbuf) < 0)
			continue;	/* Doesn't exist */
		q = (struct queue *)malloc(sizeof(time_t)+strlen(d->d_name)+1);
		if (q == NULL)
			goto errdone;
		q->q_time = stbuf.st_mtime;
		strcpy(q->q_name, d->d_name);
		/*
		 * Check to make sure the array has space left and
		 * realloc the maximum size.
		 */
		if (++nitems > arraysz) {
			queue = (struct queue **)realloc((char *)queue,
				(stbuf.st_size/12) * sizeof(struct queue *));
			if (queue == NULL)
				goto errdone;
		}
		queue[nitems-1] = q;
	}
	closedir(dirp);
	if (nitems)
		qsort(queue, nitems, sizeof(struct queue *), compar);
	*namelist = queue;
	return(nitems);

errdone:
	closedir(dirp);
	return(-1);
}

/*
 * Compare modification times.
 */
static int compar(register struct queue **p1, register struct queue **p2)
{
	if ((*p1)->q_time < (*p2)->q_time)
		return(-1);
	if ((*p1)->q_time > (*p2)->q_time)
		return(1);
	return(0);
}

/*
 * Figure out whether the local machine is the same
 * as the remote machine (RM) entry (if it exists).
 */
int checkremote(char *RemoteMachine)
{
	char name[MAXHOSTNAMELEN];
	register struct hostent *hp;
	static char errbuf[128];

	if (RemoteMachine != (char *)NULL) {
		/* get the official name of the local host */
		gethostname(name, sizeof(name));
		name[sizeof(name)-1] = '\0';
		hp = gethostbyname(name);
		if (hp == (struct hostent *) NULL) 
		{
		    sprintf(errbuf,
					"Unable to get official name for local machine (%s)\n",
					name);
			warning(errbuf);
		    return 0;
		} else (void) strcpy(name, hp->h_name);

		/* get the official name of RM */
		hp = gethostbyname(RemoteMachine);
		if (hp == (struct hostent *) NULL) 
		{
		    sprintf(errbuf,	
					"Unable to get official name for remote machine (%s)\n",
					RemoteMachine);
			warning(errbuf);
		    return 0;
		}
		/*
		 * if the two hosts are not the same,
		 * then the printer must be remote.
		 */
		if (strcmp(name, hp->h_name) != 0)
			return 1;
	}
	return 0;
}

/*
 * Get a list pointer an insert the data in it, one line per list entry
 *  If data does no end in \n, future calls will append data to the
 *  previous entry.
 */
void ParseData(List *Listing, char *data, int i)
{
	static int FlagNewLine = 1;
	int pos=0;
	static char superstring[200];
	while(i--)
	{
		if(data[pos] == '\n')
		{
			char tmp[200];
			strncpy(tmp, data, pos);
			tmp[pos] = '\0';
			if (!FlagNewLine)
			{
				strcat(superstring, tmp);
			}
			else
			{	
				superstring[0] = '\0';
				strcpy(superstring, tmp);
			}
			data = data + pos + 1;
			pos = 0;
			FlagNewLine = 1;			
			if(!NewListEntry(Listing, superstring))
			{
				fatal("Not enough memory!\n");
				exit(1);
			}
		}
		else
			pos++;
	}
	if (pos)
	{		
		if (!FlagNewLine)
			strcat(superstring, data);
		else
		{	
			superstring[0] = '\0';
			strcpy(superstring, data);
		}
		FlagNewLine = 0;
	}
}
