/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "../List.h"

#include "lp.h"
#include "error.h"
#include "common.h"
#include "printcap.h"

#define OWNCOL	7		/* start of Owner column in normal */
#define SIZCOL	62		/* start of Size column in normal */

/*-------------- Public functions -------------------*/
void RetrieveQueue(char *Printer, List* Return);
/*-------------- Private functions ------------------*/
static void InspectQueue(char *Printer);
static void header();
static void inform(char *ControlFile, char *CurrentFile, int rank);
static void blankfill(register int n);
static void dump(char *nfile, char *file, int copies, int index);
static void prank(int n);
static void PrintStatusFile(char *StatusFile);
static char *GetCurrentFileName(char *LockFile, char *StatusFile);
static void PrintForeignQueue(char *RemoteMachine, char *RemotePrinter);
static void PrintLocalQueue(char *LocalPrinter, char *SpoolingDir, 
					 char *LockFile, char *StatusFile, int sendtorem);
/*-------------- Private variables ------------------*/		
List *Listing;       /* a placehold to the results */
int		col;		/* column on screen */

/*
 * Display the current state of the queue. 
 */
void RetrieveQueue(char *Printer, List* Return)
{
	Listing = Return; /* we are going to fill Return */
	InspectQueue(Printer);
}

/*  
 * InspectQueue
 */
static void InspectQueue(char *Printer)
{
	int i;
	int sendtorem;
	char *LP, *SD, *LO, *ST, *RM, *RP;
	char line[BUFSIZ];
	char pbuf[BUFSIZ/2];	/* buffer for printcap entry */
	char *bp = pbuf;		/* pointer into ebuf for pgetent() */
	
	openlog("lpd", 0, LOG_LPR);

	if ((i = pgetent(line, Printer)) < 0)
	{		
		fatal("Cannot open printer description file\n");
		return;
	}	
	else if (i == 0)
	{	  		
		fatal("Unknown printer\n");
		return;
	}	
	if ((LP = pgetstr("lp", &bp)) == NULL)
		LP = _PATH_DEFDEVLP;
	if ((RP = pgetstr("rp", &bp)) == NULL)
		RP = DEFLP;
	if ((SD = pgetstr("sd", &bp)) == NULL)
		SD = _PATH_DEFSPOOL;
	if ((LO = pgetstr("lo", &bp)) == NULL)
		LO = DEFLOCK;
	if ((ST = pgetstr("st", &bp)) == NULL)
		ST = DEFSTAT;
	RM = pgetstr("rm", &bp);
	sendtorem = checkremote(RM);
		
	/*
	 * Print out local queue
	 * Find all the control files in the spooling directory
	 */
	PrintLocalQueue(LP, SD, LO, ST, sendtorem);
	/*
	 * Print foreign queue
	 * Note that a file in transit may show up in either queue.
	 */
	if(sendtorem)
		PrintForeignQueue(RM, RP);
}

/*  
 * PrintLocalQueue
 * 
 */
static void PrintLocalQueue(char *LocalPrinter, char *SpoolingDir, 
					 char *LockFile, char *StatusFile, int sendtorem)
{
	int nitems, i;
	struct queue **queue;
	struct queue *q;
	struct stat statb;
	char *cf;
	  
	if (chdir(SpoolingDir) < 0)
	{	  		
		fatal("Cannot chdir to spooling directory\n");
		return ;
	}	
	if ((nitems = getq(&queue, SpoolingDir)) < 0)
	{		
		fatal("Cannot examine spooling area\n");
		return ;
	}	
	if (stat(LockFile, &statb) >= 0) /* lock file */
	{
		if (statb.st_mode & 0100)  /* execute by owner -> down */
		{
			warning("Printer is down\n");
			/* status file -> why is it down? */
			PrintStatusFile(StatusFile);
		}
		if (statb.st_mode & 010) /* execute by group -> disabled */
			warning("Printer queue is turned off\n");
	}

	/* if nitems>0, we have files to print */
	if (nitems) 
	{
		int rank;
		
		cf = GetCurrentFileName(LockFile, StatusFile);
		/*
		 * Now, examine the control files and print out the jobs to
		 * be done for each user.
		 */
		rank = 0;
		header();
		for (i = 0; i < nitems; i++) 
		{
			q = queue[i];
 			if (sendtorem || strcmp(q->q_name, cf))
				rank++;
			inform(q->q_name, cf, rank);
			free(q);
		}
		free(queue);
	}
	else
	  if(!sendtorem)
		ParseData(Listing, "no entries\n", 11);
}

static void inform(char *ControlFile, char *CurrentFile, int rank)
{
	int copies, index = 0;
	long totsize = 0;
	FILE *cfp;
	char msg[MAXPATHLEN], *line, file[132];
	struct stat lbuf;

	/*
	 * There's a chance the control file has gone away
	 * in the meantime; if this is the case just keep going
	 */
	if ((cfp = fopen(ControlFile, "r")) == NULL)
	{  
		warning("Cannot read control file\n");
		return;
	}	
	copies = 0;
	while ((line = getline(cfp))) {
		switch (line[0]) {
		case 'P': /* Person *//* Was this file specified in the user's list? */
			col = 0;
			prank(rank);
			blankfill(OWNCOL);			
			sprintf(msg,"%-10s %-3d  ", line+1, atoi(ControlFile+3));
			ParseData(Listing, msg, strlen(msg));
			col += 16;
			index = 0;
			continue;
		default: /* some format specifer and file name? */
			if (line[0] < 'a' || line[0] > 'z')
				continue;
			if (copies == 0 || strcmp(file, line+1) != 0)
				strcpy(file, line+1);
			copies++;
			continue;
		case 'N':/* Name */
			if (strcmp(line+1, " ") == 0)
				dump("(standard input)", file, copies, index);
			else
				dump(line+1, file, copies, index);					
			if (*file && !stat(file, &lbuf))
				totsize += copies * lbuf.st_size;
			index++;
			file[0] = '\0';
			copies = 0;
		}
	}
	fclose(cfp);
	blankfill(SIZCOL);
	sprintf(msg, "%ld bytes\n", totsize);
	ParseData(Listing, msg, strlen(msg));
}

/*
 * Give the abbreviated dump of the file names
 */
static void dump(char *nfile, char *file, int copies, int index)
{
	short n, fill;

	/*
	 * Print as many files as will fit
	 *  (leaving room for the total size)
	 */
	fill = (index==0) ? 0 : 2;	/* fill space for ``, '' */
	if (((n = strlen(nfile)) + col + fill) >= SIZCOL-4) 
	{
		if (col < SIZCOL) 
		{
			ParseData(Listing, " ...", 4);
			col += 4;
			blankfill(SIZCOL);
		}
	} 
	else 
	{
		if (index)
			ParseData(Listing, ", ", 2);
		ParseData(Listing, nfile, strlen(nfile));
		col += n+fill;
	}
}


static void PrintForeignQueue(char *RemoteMachine, char *RemotePrinter)
{
	char line[BUFSIZ];
	int fd, i;
	
	(void) sprintf(line, "%c%s", 0 + '\3', RemotePrinter);
	strcat(line, "\n");
	fd = getport(RemoteMachine);
	if(IsThereError())
	  	return;
	if (fd < 0) 
	{
		char msg[BUFSIZ];
		sprintf("Connection to %s is down\n", RemoteMachine);
		fatal(msg);
		return;
	}
	else 
	{
		i = strlen(line);
		if (write(fd, line, i) != i)
		{  			
			fatal("Lost connection\n");
			return;
		}		
		while ((i = read(fd, line, sizeof(line))) > 0)
			ParseData(Listing, line, i);
		(void) close(fd);
	}
}

/*  
 * GetCurrentFileName
 */
static char *GetCurrentFileName(char *LockFile, char *StatusFile)
{
	FILE *fp;
	char *cp;
	int i;
	static char current[40];
	
	fp = fopen(LockFile, "r");
	if (fp == NULL)
	{
		warning("No daemon present\n");
		current[0] = '\0';
	}
	else 
	{
		/* get daemon pid */
		/* fill current in */
		cp = current;
		while ((*cp = getc(fp)) != EOF && *cp != '\n')
			cp++;
		*cp = '\0';
		i = atoi(current);
		/* i contains daemon pid */
		if (i <= 0 || kill(i, 0) < 0)
		{
			warning("No daemon present\n");
			current[0] = '\0';
		}
		else 
		{
			/* read current file name */
			cp = current;
			while ((*cp = getc(fp)) != EOF && *cp != '\n')
				cp++;
			*cp = '\0';
			PrintStatusFile(StatusFile);
		}
		fclose(fp);
	}
	return current;
}

/*
 * Print the status file.
 */
static void PrintStatusFile(char *StatusFile)
{
	int i, fd;
	char line[BUFSIZ];

 	fd = open(StatusFile, O_RDONLY);
	if (fd >= 0) 
	{
		(void) flock(fd, LOCK_SH);
		while ((i = read(fd, line, sizeof(line))) > 0)
			ParseData(Listing, line, i);
		(void) close(fd);	/* unlocks as well */
	} 
	else
		ParseData(Listing, "\n", 1);
}

/*
 * Print the header for the short listing format
 */
static void header()
{
	char	*head0 = "Rank   Owner      Job  Files";
	char	*head1 = "Total Size\n";

	ParseData(Listing, head0, strlen(head0));
	col = strlen(head0)+1;
	blankfill(SIZCOL);
	ParseData(Listing, head1, strlen(head1));
}


 /*
 * Fill the line with blanks to the specified column
 */
static void blankfill(n)
	register int n;
{
	while (col++ < n)
		ParseData(Listing, " ", 1);
}

/*
 * Print the job's rank in the queue,
 *   update col for screen management
 */
static void prank(int n)
{
	char line[100];
	static char *r[] = {
		"th", "st", "nd", "rd", "th", "th", "th", "th", "th", "th"
	};

	if (n == 0) {
		ParseData(Listing, "active", 6);
		col += 6;
		return;
	}
	if ((n/10)%10 == 1)
		(void) sprintf(line, "%dth", n);
	else
		(void) sprintf(line, "%d%s", n, r[n%10]);
	col += strlen(line);
	ParseData(Listing, line, strlen(line));
}
